"""storage.py

Storage access

created 25-jan-2020 by richb@instantlinux.net
"""

from b2sdk.v1 import B2Api, InMemoryAccountInfo
import base64
import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from datetime import datetime
from flask import g
import json
import logging
import re
import redis
from sqlalchemy.orm.exc import NoResultFound
import uuid

import apicrud.access as access
import apicrud.constants as constants
import apicrud.grants as grants
import apicrud.utils as utils


class StorageAPI(object):

    def __init__(self, credential_ttl=None, redis_conn=None, redis_host=None,
                 redis_port=6379, cache_ttl=1200, uid=None, models=None,
                 db_session=None):
        self.models = models
        self.redis_conn = (
            redis_conn or redis.Redis(host=redis_host, port=redis_port, db=0))
        self.cache_ttl = cache_ttl
        self.credential_ttl = credential_ttl or 86400
        self.uid = uid or access.AccessControl().uid
        self.db = db_session or g.db

    def get_upload_url(self, body):
        storage_id = body.get('storage_id')
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            return dict(id=storage_id, message='storage volume not found'), 404
        resource = body.get('parent_type', 'album')
        if re.compile('^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}'
                      '-?[89ab][a-f0-9]{3}-?[a-f0-9]{12}.jpeg').match(
                          body.get('name').lower()):
            # Make iPhone filename shorter
            body['name'] = base64.urlsafe_b64encode(
                uuid.UUID(body['name'].strip('.jpeg')).bytes).rstrip(
                    b'=').decode('ascii') + '.jpeg'
        logmsg = dict(uid=self.uid, resource=resource, name=body.get('name'),
                      size=body.get('size'))

        model = dict(album=self.models.Album, list=self.models.List,
                     message=self.models.Message)[resource]
        if resource == 'album':
            try:
                record = self.db.query(model).filter_by(
                    id=body.get('parent_id'), status='active')
                album_size = len(record.one().pictures)
            except NoResultFound:
                return dict(album_id=body.get(
                    'parent_id'), message='album not found'), 404
            max_size = grants.Grants(self.models,
                                     db_session=self.db,
                                     ttl=self.cache_ttl).get('album_size')
            if album_size >= max_size:
                msg = 'album is full (max=%d)' % max_size
                logging.info(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        if not access.AccessControl(
                models=self.models, model=model).with_permission(
                    'u', query=record):
            msg = 'access_denied'
            logging.warning(dict(message=msg, **logmsg))
            return dict(message='access denied'), 403

        self.vendor = storage.credentials.vendor
        id = utils.gen_id(prefix='f-')
        if body.get('content_type'):
            ctype = body['content_type'].split('/')[1]
        else:
            ctype = None
        storage_path = ('%s/%s/%s' % (
            storage.prefix if storage.prefix else '', self.uid, id)).strip('/')
        suffix = '.' + ctype if ctype else ''
        if ctype in constants.MIME_VIDEO_TYPES:
            duration_max = grants.Grants(
                self.models, db_session=self.db,
                ttl=self.cache_ttl).get('video_duration_max')
            if body.get('duration') and body['duration'] > duration_max:
                msg = 'video exceeds maximum duration=%f' % duration_max
                logging.warning(dict(message=msg, **logmsg))
                return dict(message=msg), 405
        max_size = grants.Grants(
            self.models, db_session=self.db,
            ttl=self.cache_ttl).get('media_size_max')
        if body.get('size') and body['size'] > max_size:
            msg = 'file size exceeds max=%d' % max_size
            logging.info(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        if self.vendor == 'backblaze':
            result, status = StorageBackblaze(
                self.models.Storage).get_upload_url(
                    storage_id, storage_path + suffix,
                    content_type=body.get('content_type'))
        elif self.vendor == 'aws':
            result, status = StorageS3(
                credentials=storage.credentials,
                credential_ttl=self.credential_ttl).get_upload_url(
                    storage.bucket, storage_path + suffix,
                    content_type=body.get('content_type'))
        else:
            return dict(message='Unsupported vendor=%s' % self.vendor), 405
        if status != 201:
            return result, status
        meta = dict(
            created=g.request_start_time.timestamp(),
            ctype=ctype,
            duration=body.get('duration'),
            fid=id,
            height=body.get('height'),
            modified=body.get('modified'),
            name=body.get('name'),
            oid=storage.uid,
            path=storage_path,
            pid=body.get('parent_id'),
            sid=storage_id,
            size=body.get('size'),
            uid=self.uid,
            width=body.get('width'))
        self.update_file_meta(id, meta)
        logging.info(dict(action='get_upload_url', **logmsg))
        return dict(params=result['fields'], upload_url=result['uploadUrl'],
                    file_id=id), status

    def get_file_meta(self, file_id):
        """
        returns:
        dict with fid, uid, sid, pid, path, name
        """
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            meta = json.loads(self.redis_conn.get(key))
        except TypeError:
            return None
        except Exception as ex:
            logging.warn('action=get_upload_state uid=%s exception=%s' %
                         (self.uid, str(ex)))
            return None
        return meta

    def update_file_meta(self, file_id, meta):
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            self.redis_conn.set(key, json.dumps(meta), ex=self.cache_ttl,
                                nx=True)
        except Exception as ex:
            logging.warn('action=update_file_meta uid=%s exception=%s' %
                         (self.uid, str(ex)))

    def del_file_meta(self, file_id, meta):
        key = 'fid:%s:%s' % (self.uid, file_id[-8:])
        try:
            self.redis_conn.delete(key)
        except Exception as ex:
            logging.warn('action=del_file_meta uid=%s exception=%s' %
                         (self.uid, str(ex)))

    def upload_complete(self, file_id, status, func_worker):
        """ Finalize upload - pass to the worker

        params:
          file_id - ID of file meta in redis
          status - completion status from frontend ('done' if OK)
          func_worker - callback function for passing to celery worker
        """

        meta = self.get_file_meta(file_id)
        if meta and status == 'done':
            duration = datetime.utcnow().timestamp() - meta['created']
            logging.info(dict(action='upload_complete', file_id=file_id,
                              duration='%.3f' % duration))
            func_worker(self.uid, file_id)
            return dict(file_id=file_id), 201
        else:
            msg = 'unhandled frontend status'
            logging.warning('action=upload_complete status=%s message=%s' %
                            (status, msg))
            return dict(file_id=file_id, message=msg), 405

    def fetch_album_meta(self, album_id, thumbnail_height):
        """Retrieve metadata of a media album in format suitable
        for react-image-gallery
        """

        # TODO table-join to pick up rank of each picture
        album = self.db.query(self.models.Album).filter_by(id=album_id).one()
        results = []
        for picture in album.pictures:
            if picture.status != 'active':
                continue
            orig_uri = None
            uri_path = '%s/%s' % (picture.storage.cdn_uri, picture.path)
            fmt = picture.format_original
            sizes = [int(x) for x in album.sizes.split(',')
                     if not picture.height or int(x) <= picture.height]
            if fmt in constants.MIME_VIDEO_TYPES:
                # TODO see line 249 of xiaolin's app.js example
                logging.warning('resource=album action=get id=%s '
                                'message=videos_nyi' % picture.id)
                continue
            else:
                max_height = grants.Grants(
                    self.models, self.db,
                    ttl=self.cache_ttl).get('photo_res_max')
                if not picture.height or picture.height <= max_height:
                    orig_uri = '%s.%s' % (uri_path, fmt)
                else:
                    orig_uri = '%s.%d.%s' % (uri_path, max_height, fmt)
            results.append(dict(
                id=picture.id,
                description=picture.caption,
                duration=picture.duration,
                height=picture.height,
                imageSet=[dict(srcSet='%s.%d.%s' % (uri_path, res, fmt),
                               media='(max-width: %dpx)' % res) for
                          res in sizes],
                name=picture.name,
                original=orig_uri,
                path=picture.path,
                size=picture.size,
                thumbnail='%s.%d.%s' % (uri_path, thumbnail_height, fmt),
                type='%s/%s' % ('video' if fmt in constants.MIME_VIDEO_TYPES
                                else 'image', fmt),
                uid=picture.uid,
                width=picture.width))
            if picture.height and picture.height <= max_height:
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%s' % (uri_path, fmt),
                    media='(max-width: %dpx)' % picture.height))
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%s' % (uri_path, fmt),
                    media='(min-width: %dpx)' % picture.height))
            else:
                results[-1]['imageSet'].append(dict(
                    srcSet='%s.%d.%s' % (uri_path, sizes[-1], fmt),
                    media='(min-width: %dpx)' % sizes[-1]))
        return results

    def get_object(self, storage_id, storage_key):
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).get_object(
                                 storage.bucket, storage_key)

    def put_object(self, storage_id, storage_key, content,
                   content_type='image/jpeg'):
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).put_object(
                                 storage.bucket, storage_key, content,
                                 content_type=content_type)

    def del_object(self, storage_id, storage_key):
        try:
            storage = self.db.query(self.models.Storage).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            logging.error('storage volume not found')
        if storage.credentials.vendor == 'aws':
            return StorageS3(credentials=storage.credentials).del_object(
                                 storage.bucket, storage_key)


class StorageBackblaze(object):

    def __init__(self, model, db_session=None):
        self.api = B2Api(InMemoryAccountInfo())
        self.db = db_session
        self.model = model

    def get_upload_url(self, storage_id, storage_key,
                       content_type='image/jpeg'):
        try:
            storage = self.db.query(self.model).filter_by(
                id=storage_id, status='active').one()
        except NoResultFound:
            return dict(id=storage_id, message='storage volume not found'), 404

        self.api.authorize_account(
            'production', storage.credentials.key, storage.credentials.secret)
        try:
            result = self.api.raw_api.get_upload_url(
                self.api.account_info.get_api_url(),
                self.api.account_info.get_account_auth_token(),
                storage.identifier)
        except Exception as ex:
            msg = str(ex)
            return dict(message=msg, id=storage_id), 405
        return result, 201


class StorageS3(object):

    def __init__(self, credential_ttl=3600, credentials=None, access_key=None,
                 secret_key=None, region=constants.DEFAULT_AWS_REGION):
        self.credential_ttl = credential_ttl
        self.region = region
        if credentials:
            self.access_key = credentials.key
            self.secret_key = credentials.secret
        else:
            self.access_key = access_key
            self.secret_key = secret_key
        self.api = boto3.client(
            's3', aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name=self.region,
            # signature_version = 's3v4',
            config=Config(s3={'addressing_style': 'virtual'}))

    def get_upload_url(self, bucket, storage_key,
                       content_type='image/jpeg'):
        try:
            result = self.api.generate_presigned_post(
                bucket, storage_key, ExpiresIn=self.credential_ttl)
        except ClientError as ex:
            msg = str(ex)
            return dict(message=msg, bucket=bucket), 405
        return dict(fields=result['fields'], uploadUrl=result['url']), 201

    def get_object(self, bucket, storage_key):
        """Fetch the object into a byte array"""

        return self.api.get_object(Bucket=bucket,
                                   Key=storage_key)['Body'].read()

    def put_object(self, bucket, storage_key, content,
                   content_type='image/jpeg'):
        """Store a byte-array object"""

        logging.debug(dict(action='put', bucket=bucket,
                           storage_key=storage_key, content_type=content_type))
        return self.api.put_object(
            Bucket=bucket, Key=storage_key, Body=content,
            ContentType=content_type)

    def del_object(self, bucket, storage_key):
        """Delete an object"""

        logging.debug(dict(action='del', bucket=bucket,
                           storage_key=storage_key))
        return self.api.delete_object(Bucket=bucket, Key=storage_key)



"""service_registry.py

Services or the UI discover one another through this service registry.
Each microservice instance submits its identity and capabilities to
this central registry, implemented as expiring redis keys which are
updated at a fixed frequency. Encryption provides modest protection
against injection attacks.

created 8-may-2020 by richb@instantlinux.net

"""

import json
import logging
import redis
import socket
import threading

from .aes_encrypt import AESEncrypt
from .utils import utcnow

refresh_thread = None
service_data = {}
params = {}


class ServiceRegistry(object):

    def __init__(self, config, ttl=None, redis_conn=None):
        global params, refresh_thread
        self.config = config
        if not redis_conn and hasattr(config, 'redis_conn'):
            redis_conn = config.redis_conn
        if not params:
            params = dict(
                aes=AESEncrypt(self.config.REDIS_AES_SECRET),
                connection=(
                    redis_conn or redis.Redis(host=config.REDIS_HOST,
                                              port=config.REDIS_PORT, db=0)),
                interval=config.REGISTRY_INTERVAL,
                ttl=ttl or config.REGISTRY_TTL)
        if not refresh_thread:
            refresh_thread = threading.Thread()

    def register(self, resource_endpoints, service_name=None,
                 instance_id=socket.gethostname(), tcp_port=None):
        global service_data
        service_data['info'] = dict(
            endpoints=resource_endpoints,
            ipv4=socket.gethostbyname(socket.gethostname()),
            port=tcp_port or self.config.PORT,
            public_url=self.config.PUBLIC_URL,
            created=utcnow().replace(microsecond=0).isoformat())
        service_data['name'] = service_name or self.config.SERVICE_NAME
        service_data['id'] = instance_id
        refresh_thread = threading.Timer(1, ServiceRegistry.update, ())
        refresh_thread.start()

    @staticmethod
    def update():
        key = 'reg:%s:%s' % (service_data['name'], service_data['id'])
        try:
            params['connection'].set(key, params['aes'].encrypt(json.dumps(
                service_data['info'])), ex=params['ttl'], nx=False)
        except Exception as ex:
            logging.error('action=registry.update message=%s' % str(ex))
        refresh_thread = threading.Timer(params['interval'],
                                         ServiceRegistry.update, ())
        refresh_thread.start()

    def find(self, service_name=None):
        """ Finds one or all services

        returns:
          dict - instances (list of registered services)
                 url_map (public url for each top-level resource)
        """

        key = 'reg:%s:*' % service_name if service_name else 'reg:*'
        retval = []
        url_map = {}
        for instance in params['connection'].scan_iter(match=key):
            try:
                info = json.loads(params['aes'].decrypt(
                    params['connection'].get(instance)))
                info['name'], info['id'] = instance.decode().split(':')[1:]
                retval.append(info)
            except TypeError:
                # A decrypt failure probably means another cluster
                # is sharing the same redis instance (e.g. in a dev env)
                # so skip over those service instances
                continue
            except Exception as ex:
                logging.warn('action=registry.find key=%s exception=%s' %
                             (instance, str(ex)))
            if not info.get('endpoints'):
                continue
            logging.info(dict(step=1, instance=instance, **info))
            for resource in info['endpoints']:
                if resource not in url_map:
                    url_map[resource] = info[
                        'public_url'] + self.config.BASE_URL
                elif url_map[resource] != info[
                        'public_url'] + self.config.BASE_URL:
                    logging.warning(dict(
                        action='registry.find',
                        message='instance serves different public URL',
                        resource=resource, instance1=instance.decode(),
                        url1=info['public_url'], url2=url_map[resource]))
        return dict(instances=retval, url_map=url_map)



"""constants.py

Constants for API library

Values that can be modified by a container restart belong in config.py

created 31-mar-2019 by richb@instantlinux.net
"""

AUTH_INVITEE = 'invitee'
DEFAULT_AWS_REGION = 'us-east-2'
LIB_MOD_SPATIALITE = ['/usr/lib/x86_64-linux-gnu/mod_spatialite.so',
                      '/usr/lib/mod_spatialite.so.7']
MIME_IMAGE_TYPES = ("gif", "heic", "jpeg", "png", "svg")
MIME_VIDEO_TYPES = ("mp4", "mpeg")
PER_PAGE_DEFAULT = 100
REDIS_TTL = 3600


class i18n:
    # These are keys in a user-provided dict; see the
    # example/i18n_textstrings.py for how to assign string
    # templates for messaging
    # TODO - update this doc when messaging code is moved
    # into this library
    CONTACT_ADDED_REQUEST = 'contact_add'
    PASSWORD_RESET = 'password_reset'



"""session_manager.py

Each session is stored as an encrypted JSON dict in redis, indexed
by sub:token

created 8-may-2019 by richb@instantlinux.net
"""

from datetime import datetime, timedelta
import json
import logging
import random
import redis
import string
import time

from . import constants
from .aes_encrypt import AESEncrypt


class SessionManager(object):

    def __init__(self, config, ttl=constants.REDIS_TTL, redis_conn=None):
        self.config = config
        self.connection = (
            redis_conn or redis.Redis(host=config.REDIS_HOST,
                                      port=config.REDIS_PORT, db=0))
        self.ttl = ttl
        self.aes = AESEncrypt(self.config.REDIS_AES_SECRET)

    def create(self, uid, roles, **kwargs):
        """create a session, which is an encrypted JSON object with the values
        defined in https://tools.ietf.org/html/rfc7519 for JWT claim names:
        - exp - expiration time, as integer Unix epoch time
        - iss - a constant JWT_ISSUER
        - jti - JWT ID, the randomly-generated token
        - sub - the uid of a user

        We add these:
        - auth - authorized roles
        - any other key=value pairs the caller passes as kwargs

        The session automatically expires based on object's ttl. Part of
        the jti token is used in redis key, to allow a user to log into
        multiple sessions. The rest of the token is encrypted, to
        secure it from replay attack in the event redis traffic
        is compromised.
        """

        token = kwargs.pop('nonce', _gen_id(prefix=''))
        ttl = kwargs.pop('ttl', self.ttl)
        params = dict(
            auth=':'.join(roles),
            exp=(datetime.utcnow() + timedelta(
                seconds=ttl)).strftime('%s'),
            iss=self.config.JWT_ISSUER, jti=token, sub=uid
        )
        key = 'uid:%s:%s' % (uid, token[-3:])
        content = {**params, **kwargs}  # noqa
        if self.connection.set(key, self.aes.encrypt(json.dumps(
                content)), ex=ttl, nx=True):
            return content

    def get(self, uid, token, arg=None):
        key = 'uid:%s:%s' % (uid, token[-3:])
        try:
            data = json.loads(self.aes.decrypt(
                self.connection.get(key)))
        except TypeError:
            return None
        except Exception as ex:
            logging.info('action=session.get uid=%s exception=%s' %
                         (uid, str(ex)))
            return None
        if data['jti'] != token:
            logging.warning('action=session.get message=rejected')
            return None
        if arg:
            return data[arg] if arg in data else None
        return data

    def update(self, uid, token, arg, value):
        key = 'uid:%s:%s' % (uid, token[-3:])
        data = self.get(uid, token)
        data[arg] = value
        self.connection.set(key, self.aes.encrypt(
            json.dumps(data)), ex=self.ttl)

    def delete(self, uid, token):
        self.connection.delete('uid:%s:%s' % (uid, token[-3:]))


class Mutex:
    """Simple mutex implementation for non-clustered Redis

    Tried to implement this as inner class to DRY out the init,
    but ... no joy.
    """
    def __init__(self, lockname, redis_host=None, maxwait=20,
                 ttl=constants.REDIS_TTL, redis_conn=None):
        self.connection = (
            redis_conn or redis.Redis(host=redis_host, port=6379, db=0))
        self.ttl = ttl
        self.lock_signature = _gen_id()
        self.lockname = lockname
        self.maxwait = maxwait

    def acquire(self):
        for retry in range(self.maxwait):
            if self.connection.set('mutex:%s' % self.lockname,
                                   self.lock_signature, ex=self.ttl, nx=True):
                return True
            time.sleep(1)
        raise TimeoutError('Unable to acquire mutex=%s' % self.lockname)

    def release(self):
        if ((self.connection.get('mutex:%s' % self.lockname).decode('utf-8') !=
             self.lock_signature) or
                not self.connection.delete('mutex:%s' % self.lockname)):
            raise TimeoutError('Mutex held too long, may need a longer ttl')

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.release()


# TODO: resolve another DRY problem - this yields a circular dependency
# when invoked as 'import lib.utils'
def _gen_id(length=8, prefix='x-', chars=(
        '-' + string.digits + string.ascii_uppercase + '_' +
        string.ascii_lowercase)):
    def _int2base(x, chars, base=64):
        return _int2base(x // base, chars, base).lstrip(chars[0]) + chars[
            x % base] if x else chars[0]
    return (prefix +
            _int2base((datetime.utcnow() - datetime(2018, 1, 1)).days * 8 +
                      random.randint(0, 8), chars) +
            ''.join(random.choice(chars) for i in range(length - 3)))



"""aes_encrypt.py

Provides easier-to-use AES CBC encrypt/decrypt operations for strings

created 14-may-2019 by richb@instantlinux.net
"""

import base64
import hashlib

from Cryptodome.Cipher import AES
from Cryptodome.Random import get_random_bytes

BLOCK_SIZE = 16


class AESEncrypt(object):
    def __init__(self, secret):
        self.private_key = hashlib.sha256(secret.encode('utf-8')).digest()

    def encrypt(self, raw):
        iv = get_random_bytes(AES.block_size)
        cipher = AES.new(self.private_key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(self._pad(raw).encode()))

    def decrypt(self, enc):
        enc = base64.b64decode(enc)
        iv = enc[:16]
        cipher = AES.new(self.private_key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[16:]))

    @staticmethod
    def _pad(s):
        return s + (BLOCK_SIZE - len(s) % BLOCK_SIZE) * chr(
            BLOCK_SIZE - len(s) % BLOCK_SIZE)

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s) - 1:])]



"""health.py

created 28-mar-2020 by richb@instantlinux.net
"""

from flask import g


def healthcheck(app_name='api', service_name='main', tests=None, model=None,
                releaseId=None, build_date=None, version=None):
    """Returns a health check report, with optional tests, in
    format recommended by Nadareishvili in RFC draft
    https://tools.ietf.org/id/draft-inadarei-api-health-check-04.html

    params:
      service_name - microservice name for serviceId
      tests - optional tests to run
      model - schema model (usually AlembicVersion)
    """

    status = 200
    retval = dict(
        description="%s - %s" % (app_name, service_name),
        notes=["build_date:%s" % build_date],
        serviceId=service_name)
    if releaseId:
        retval['releaseId'] = releaseId
    if version:
        retval['version'] = version
    if model:
        try:
            retval['notes'].append(
                'schema:%s' % g.db.query(model).one().version_num)
        except Exception as ex:
            retval['output'] = str(ex)
            status = 503
    return dict(
        status="pass" if status == 200 else "fail", **retval), status



"""basic_crud.py

Basic CRUD
  Base object for create/read/update/delete/find controller operations.


created 31-mar-2019 by richb@instantlinux.net
"""

import base64
from connexion import NoContent
from datetime import datetime, timedelta
from flask import g, request
import json
import logging
from sqlalchemy import asc, desc
from sqlalchemy.exc import IntegrityError, InvalidRequestError
from sqlalchemy.orm.exc import NoResultFound

from . import constants
from .access import AccessControl
from .account_settings import AccountSettings
from . import geocode, singletons, utils


class BasicCRUD(object):

    def __init__(self, config=None, models=None, resource=None, model=None):
        self.config = config
        self.models = models
        self.resource = resource
        if self.resource not in singletons.controller:
            self.model = model
            singletons.controller[self.resource] = self

    @staticmethod
    def create(body, id_prefix='x-'):
        self = singletons.controller[request.url_rule.rule.split('/')[3]]
        acc = AccessControl(models=self.models, model=self.model)
        logmsg = dict(action='create', account_id=acc.account_id,
                      resource=self.resource, ident=acc.identity)
        if 'id' in body:
            return dict(message='id is a read-only property',
                        title='Bad Request'), 405
        body['id'] = utils.gen_id(prefix=id_prefix)
        body['created'] = utils.utcnow()
        if body.get('expires'):
            try:
                body['expires'] = datetime.strptime(
                    body['expires'], '%Y-%m-%dT%H:%M:%S.%fZ')
            except ValueError:
                body['expires'] = datetime.strptime(
                    body['expires'], '%Y-%m-%dT%H:%M:%SZ')
        if hasattr(self.model, 'uid') and not body.get('uid'):
            body['uid'] = acc.uid
        uid = body.get('uid')
        if not acc.with_permission('c', new_uid=uid,
                                   membership=acc.primary_res,
                                   id=body.get('event_id')):
            logging.warning(dict(
                message='access denied', uid=uid, **logmsg))
            return dict(message='access denied'), 403
        logmsg['uid'] = body.get('uid')
        logging.info(dict(id=body['id'], name=body.get('name'), **logmsg))
        if not body.get('category_id') and hasattr(self.model, 'category_id'):
            if acc.account_id:
                body['category_id'] = AccountSettings(
                    acc.account_id, self.config, self.models,
                    g.db).get.category_id
            elif body.get('event_id'):
                body['category_id'] = g.db.query(self.models.Event).filter_by(
                    id=body['event_id']).one().category_id
            else:
                logging.warning(dict(message='unexpected no creds', **logmsg))
                return dict(message='access denied'), 403
        if hasattr(self.model, 'status'):
            body['status'] = body.get('status', 'active')
        try:
            record = self.model(**body)
        except (AttributeError, TypeError) as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message=str(ex)), 405
        g.db.add(record)
        try:
            g.db.commit()
        except IntegrityError as ex:
            message = 'duplicate or other conflict'
            logging.warning(dict(message=message, error=str(ex), **logmsg))
            return dict(message=message, data=str(body)), 405
        except Exception as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message=str(ex), data=str(body)), 405
        return dict(id=body['id']), 201

    @staticmethod
    def get(id):
        self = singletons.controller[request.url_rule.rule.split('/')[3]]
        acc = AccessControl(models=self.models, model=self.model)
        try:
            query = self.db_get(id)
            record = query.one()
        except NoResultFound as ex:
            return dict(id=id, message=str(ex)), 404

        retval = record.as_dict()
        if hasattr(self.model, 'uid') and hasattr(self.model, 'owner'):
            retval['owner'] = record.owner.name
        if hasattr(self.model, 'category_id') and retval['category_id']:
            retval['category'] = record.category.name
        if 'modified' in retval and not retval['modified']:
            del(retval['modified'])
        eid = None if hasattr(self.model, 'event_id') else (
            acc.auth_ids[acc.primary_res] or [None])[0]
        retval['rbac'] = ''.join(sorted(list(
            acc.rbac_permissions(query=query, membership=acc.primary_res,
                                 id=eid) - set('c'))))
        if 'geolat' in retval:
            access = 'r' if 'r' in retval['rbac'] else None
            if retval['privacy'] == 'secret' and not access:
                return dict(message='access denied'), 403
            return geocode.with_privacy(retval, access), 200
        if 'r' in retval['rbac']:
            logging.info(dict(
                action='get', resource=self.resource, id=id,
                ident=acc.identity, duration=utils.req_duration()))
            return retval, 200
        else:
            return dict(message='access denied', id=id), 403

    @staticmethod
    def update(id, body, access='u'):
        if 'id' in body and body['id'] != id:
            return dict(message='id is a read-only property',
                        title='Bad Request'), 405
        self = singletons.controller[request.url_rule.rule.split('/')[3]]
        body['modified'] = utils.utcnow()
        if body.get('expires'):
            try:
                body['expires'] = datetime.strptime(
                    body['expires'], '%Y-%m-%dT%H:%M:%S.%fZ')
            except ValueError:
                body['expires'] = datetime.strptime(
                    body['expires'], '%Y-%m-%dT%H:%M:%SZ')
        logmsg = dict(action='update', resource=self.resource, id=id,
                      ident=AccessControl().identity)
        try:
            query = g.db.query(self.model).filter_by(id=id)
            if not AccessControl(
                    models=self.models, model=self.model).with_permission(
                        access, query=query):
                return dict(message='access denied', id=id), 403
            current = query.one().as_dict()
            query.update(body)
            g.db.commit()
        except Exception as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message='exception %s' % str(ex)), 405

        updated = {key: body[key] for key, val in body.items()
                   if current[key] != val and key not in ('name', 'modified')}
        logging.info(dict(name=body.get('name'), duration=utils.req_duration(),
                          **logmsg, **updated))
        return dict(id=id, message='updated'), 200

    @staticmethod
    def delete(ids, force=False):
        # TODO - update auth if model could affect any session's auth
        self = singletons.controller[request.url_rule.rule.split('/')[3]]
        logmsg = dict(action='delete', resource=self.resource,
                      account_id=AccessControl().account_id,
                      ident=AccessControl().identity)
        errors = 0
        count = 0
        for id in ids:
            try:
                query = g.db.query(self.model).filter_by(id=id)
                if not AccessControl(
                        self.models, model=self.model).with_permission(
                            'd', query=query):
                    return dict(message='access denied', id=id), 403
                if force:
                    if query.delete():
                        logging.info(dict(id=id, **logmsg))
                    else:
                        logging.info(dict(id=id, msg='query failed', **logmsg))
                        errors += 1
                else:
                    logging.info(dict(id=id, status='disabled', **logmsg))
                    query.update(dict(status='disabled'))
            except Exception as ex:
                logging.warning(dict(message=str(ex), **logmsg))
                return dict(message='exception %s' % str(ex)), 405
            count += 1
        try:
            g.db.commit()
        except Exception as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message='exception %s' % str(ex)), 405
        logging.info(dict(count=count, ids=ids, **logmsg))
        return NoContent, 404 if errors else 204

    @staticmethod
    def find(**kwargs):
        """query parameters are passed from connexion by name, in
        a dictionary that also includes user and token info

        params:
        cursor_next - pagination token to fetch subsequent records
        filter (dict) - field/value pairs to query (simple queries
            only, with string or list matching; or * for any)
        limit (int) - max records to fetch
        offset (int) - old-style pagination starting offset
        sort (str) - <field>[:{asc|desc}]
        status (str) - value is added to filter

        returns:
        dict - items (list), count(int), cursor_next (str)
        """

        self = singletons.controller[request.url_rule.rule.split('/')[3]]
        acc = AccessControl(models=self.models, model=self.model)
        logmsg = dict(action='find', ident=acc.identity,
                      resource=self.resource)
        conditions = {item: value for item, value in kwargs.items()
                      if item in ('status')}
        if 'cursor_next' in kwargs:
            offset = int(self._fromb64(kwargs['cursor_next']).split(':')[1])
        elif 'offset' in kwargs:
            offset = int(kwargs['offset'])
        else:
            offset = 0
        sort = kwargs.get('sort',
                          'name' if hasattr(self.model, 'name') else 'id')
        if ':' in sort:
            sort, dir = sort.split(':')
            sortdir = desc if dir == 'desc' else asc
        else:
            sortdir = asc
        query = g.db.query(self.model)
        limit = int(kwargs.get('limit', constants.PER_PAGE_DEFAULT))
        try:
            filter = json.loads(kwargs.get('filter', '{}'))
        except json.decoder.JSONDecodeError as ex:
            msg = 'invalid filter string=%s' % str(ex)
            logging.error(dict(message=msg, **logmsg))
            return dict(message=msg), 405
        if 'id' in filter:
            # TODO: this seems a bit hackish, try to remember what
            # use-case this supports. The del() use-case is to
            # enable get:guest?filter{id:null,event_id=foo}
            if not filter.get('id'):
                if len(filter) == 1:
                    return dict(count=0, items=[]), 200
                else:
                    del(filter['id'])
        elif 'filter' in kwargs and not filter:
            return dict(count=0, items=[]), 200
        for key in filter.copy():
            if filter[key] == '*':
                filter.pop(key)
            elif (filter[key] and type(filter[key]) is str and '%'
                  in filter[key]):
                query = query.filter(getattr(self.model, key).like(
                    filter[key]))
                filter.pop(key)
            elif type(filter[key]) is list:
                items = filter.pop(key)
                if len(items) == 1:
                    filter[key] = items[0]
                else:
                    query = query.filter(getattr(self.model, key).in_(items))
        if hasattr(self.model, 'starts'):
            if filter.get('prev'):
                # TODO confirm whether to use utcnow()
                query = query.filter(self.model.starts < datetime.now())
            else:
                # Filter out records that start more than 12 hours ago
                query = query.filter(self.model.starts >
                                     datetime.now() - timedelta(hours=12))
            filter.pop('prev', None)
        try:
            query = query.filter_by(**filter).order_by(
                sortdir(getattr(self.model, sort)))
        except InvalidRequestError as ex:
            logging.warning(dict(message=str(ex), **logmsg))
            return dict(message='invalid filter specified'), 405
        if 'status' in conditions and conditions['status'] == 'disabled':
            query = query.filter(self.model.status == 'disabled')
        else:
            query = query.filter(self.model.status != 'disabled')
        query = acc.with_filter(query)
        try:
            results = query.slice(offset, offset + limit + 1).all()
        except Exception as ex:
            logging.error(dict(message=str(ex), **logmsg))
            return dict(message='Backend problem--probable bug'), 405
        retval = dict(items=[], count=query.count())
        count = 0
        for result in results[:limit]:
            record = result.as_dict()
            if hasattr(self.model, 'owner'):
                record['owner'] = result.owner.name
            # TODO - find a way to avoid this, causes KeyError on lazy-load
            if record.get('category_id'):
                record['category'] = result.category.name
            # TODO - takes 30ms per record for rbac: optimize or redesign
            record['rbac'] = ''.join(sorted(list(
                acc.rbac_permissions(owner_uid=record.get('uid'),
                                     membership=acc.primary_res,
                                     id=record.get('id'),
                                     privacy=record.get('privacy'))
                - set('c'))))
            retval['items'].append(record)
            count += 1
        if len(results) > limit:
            retval['cursor_next'] = self._tob64('cursor:%d' % (offset + limit))
        elif count < retval['count'] and count < limit and offset == 0:
            # TODO find a way to get query.count() to return accurate value
            retval['count'] = count
        logging.info(dict(
            offset=offset, limit=limit,
            duration=utils.req_duration(), **conditions, **logmsg))
        return retval, 200

    @staticmethod
    def _tob64(text):
        return base64.b64encode(bytes(text, 'utf8')).decode('ascii')

    @staticmethod
    def _fromb64(text):
        return base64.b64decode(bytes(text, 'utf8')).decode('ascii')

    def db_get(self, id):
        return g.db.query(self.model).filter_by(id=id)



"""worker_processing.py

Media worker functions to process media

created 3-feb-2020 by richb@instantlinux.net
"""

from datetime import datetime
import hashlib
import io
import logging
from PIL import Image
from PIL.ExifTags import GPSTAGS, TAGS

from ..account_settings import AccountSettings
from .storage import StorageAPI


class MediaUploadException(Exception):
    pass


class MediaProcessing(object):
    def __init__(self, uid, file_id, config, models, db_session=None):
        self.api = StorageAPI(redis_host=config.REDIS_HOST, uid=uid,
                              models=models, db_session=db_session)
        self.file_id = file_id
        self.config = config
        self.models = models
        self.meta = self.api.get_file_meta(file_id)

    def __del__(self):
        self.api.del_file_meta(self.file_id, self.meta)

    def photo(self, uid, meta, db_session):
        """metadata and scaling for still images
        """

        def _exif_get(field, max_width):
            if exif.get(field):
                return exif.get(field)[:max_width]

        # fetch the image
        suffix = "." + meta["ctype"]
        raw = self.api.get_object(meta["sid"], meta["path"] + suffix)
        if not raw:
            raise MediaUploadException("get_object failed for path=%s" %
                                       meta["path"])
        img = Image.open(io.BytesIO(raw))

        # parse its EXIF fields
        exif_raw = (hasattr(img, "_getexif") and img._getexif()) or {}
        exif = {TAGS[key]: val for key, val in exif_raw.items()
                if key in TAGS}
        geolat, geolong, altitude = None, None, None
        if exif.get("GPSInfo"):
            geolat, geolong, altitude = self._gpsexif(exif)

        # generate the 50-pixel thumbnail
        thumbnail = img.copy()
        thumbnail.thumbnail((self.config.THUMBNAIL_TINY,
                             self.config.THUMBNAIL_TINY), Image.ANTIALIAS)
        tbytes = io.BytesIO()
        thumbnail.save(tbytes, meta["ctype"])

        # construct and save a pictures db record
        try:
            if meta["modified"]:
                modified = datetime.strptime(meta["modified"],
                                             "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            logging.warn("action=photo message=invalid_date id=%s "
                         "modified=%s" % (meta["fid"], meta["modified"]))
        record = self.models.Picture(
            id=meta["fid"],
            name=meta["name"],
            path=meta["path"],
            uid=uid,
            category_id=AccountSettings(None, self.config, self.models,
                                        db_session=db_session,
                                        uid=meta["oid"]).get.category_id,
            compression=_exif_get("Compression", 16),
            datetime_original=exif.get("DateTimeOriginal") or modified,
            format_original=meta["ctype"],
            geolat=geolat,
            geolong=geolong,
            gps_altitude=altitude,
            height=img.height,
            is_encrypted=False,
            make=_exif_get("Make", 16),
            model=_exif_get("Model", 32),
            orientation=exif.get("Orientation"),
            privacy=u"invitee",
            sha1=hashlib.sha1(raw).digest(),
            sha256=hashlib.sha256(raw).digest(),
            thumbnail50x50=tbytes.getvalue(),
            size=meta["size"],
            status="active",
            storage_id=meta["sid"],
            width=img.width,
        )
        db_session.add(record)

        # add picture to album; first added is initial cover picture
        album = db_session.query(self.models.Album).filter_by(
            id=meta["pid"]).one()
        record = self.models.AlbumContent(
            album_id=meta["pid"],
            picture_id=meta["fid"],
            rank=len(album.pictures) + 1)
        db_session.add(record)
        if len(album.pictures) == 0 or album.cover_id is None:
            album.cover_id = meta["fid"]
            db_session.add(album)
        db_session.commit()

        # generate additional scaled sizes
        sizes = [int(x) for x in album.sizes.split(",")]
        for alt_size in sizes:
            if alt_size >= img.height:
                logging.info("skipping image name=%s resize=%d original=%d" %
                             (meta["name"], alt_size, img.height))
                continue
            width = int(img.width * alt_size / img.height)
            logging.info("generating image name=%s width=%d height=%d id=%s"
                         % (meta["name"], width, alt_size, meta["fid"]))
            alt_img = img.resize((width, alt_size), Image.ANTIALIAS)
            tbytes = io.BytesIO()
            alt_img.save(tbytes, meta["ctype"])
            if not self.api.put_object(
                    meta["sid"], "%s.%d%s" % (
                        meta["path"], alt_size, suffix),
                    tbytes.getvalue(),
                    content_type="image/%s" % meta["ctype"]):
                raise MediaUploadException("put_object failed")

        # stay in budget--throw away originals larger than max
        # TODO grants needs to work under celery
        # res_max = Grants(db_session=db_session,
        #                  ttl=config.CACHE_TTL).get('photo_res_max')
        res_max = self.config.DEFAULT_GRANTS.get('photo_res_max')
        if img.height > res_max:
            self.api.del_object(meta["sid"], meta["path"] + suffix)

        # TODO (eventually)-apply encryption when gallery frontend has support
        #  and strip exif tags

    def video(self, uid, meta, db_session):
        """metadata for videos - construct and save a pictures db record"""

        # TODO decide whether to bother with exif or thumbnail

        try:
            if meta["modified"]:
                modified = datetime.strptime(meta["modified"],
                                             "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            logging.warn("action=video message=invalid_date id=%s "
                         "modified=%s" % (meta["fid"], meta["modified"]))
        record = self.models.Picture(
            id=meta["fid"],
            name=meta["name"],
            path=meta["path"],
            uid=uid,
            category_id=AccountSettings(None, self.config, self.models,
                                        db_session=db_session,
                                        uid=meta["oid"]).get.category_id,
            compression=None,
            datetime_original=modified,
            duration=meta["duration"],
            format_original=meta["ctype"],
            geolat=None,
            geolong=None,
            gps_altitude=None,
            height=meta["height"],
            is_encrypted=False,
            make=None,
            model=None,
            orientation=None,
            privacy=u"invitee",
            sha1=None,
            sha256=None,
            thumbnail50x50=None,
            size=meta["size"],
            status="active",
            storage_id=meta["sid"],
            width=meta["width"],
        )
        db_session.add(record)

        # add to album
        album = db_session.query(self.models.Album).filter_by(
            id=meta["pid"]).one()
        # TODO query AlbumContents max(rank) to handle edge case
        record = self.models.AlbumContent(
            album_id=meta["pid"],
            picture_id=meta["fid"],
            rank=len(album.pictures) + 1)
        db_session.add(record)
        db_session.commit()

    def _gpsexif(self, exif):
        """Convert GPS exif data to fixed-decimal format. For some
        reason there"s no good standard python library for doing this.

        returns: tuple (lat, long, alt)
          first two are converted to fixed-precision int to fit in 32 bits
        """

        for item, val in GPSTAGS.items():
            if item in exif["GPSInfo"]:
                exif[val] = exif["GPSInfo"][item]

        # convert lat/long to fixed-precision 32-bit integers for SQL
        geolat, geolong, altitude = None, None, None
        if exif.get("GPSLatitude"):
            geolat = self._decimal_from_degrees(
                exif["GPSLatitude"], exif["GPSLatitudeRef"])
        if exif.get("GPSLongitude"):
            geolong = self._decimal_from_degrees(
                exif["GPSLongitude"], exif["GPSLongitudeRef"])
        if exif.get("GPSAltitude"):
            altitude = exif["GPSAltitude"][0] / exif["GPSAltitude"][1]
        return geolat, geolong, altitude

    @staticmethod
    def _decimal_from_degrees(dms, hemisphere):
        degrees = int((dms[0][0] / dms[0][1] +
                       dms[1][0] / dms[1][1] / 60.0 +
                       dms[2][0] / dms[2][1] / 3600.0) * 1e7)
        if hemisphere in ("S", "W"):
            degrees = -degrees
        return degrees



"""database.py

Database session management and schema update

created 31-mar-2019 by richb@instantlinux.net
"""

from flask import _app_ctx_stack
import logging
import os.path
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative import declarative_base

from . import constants
from .session_manager import Mutex

Base = declarative_base()
db_engine = None
Session = None
spatialite_loaded = False


def get_session(scopefunc=None, scoped=True, db_url=None, engine=None):
    global db_engine
    if not engine:
        # This is where init_db is invoked under celery
        engine = db_engine or _init_db(db_url=db_url)
    if scoped:
        global Session
        if scopefunc is None:
            scopefunc = _app_ctx_stack.__ident_func__
        Session = scoped_session(sessionmaker(autocommit=False,
                                              autoflush=False,
                                              bind=engine),
                                 scopefunc=scopefunc)
        return Session
    else:
        Session = sessionmaker(autocommit=False, autoflush=False, bind=engine)
        return Session


def initialize_db(models, db_url=None, engine=None, redis_conn=None,
                  redis_host=None, migrate=False, geo_support=False,
                  connection_timeout=0, schema_update=None, schema_maxtime=60):
    _init_db(db_url=db_url, engine=engine, geo_support=geo_support,
             connection_timeout=connection_timeout)
    if migrate:
        with Mutex('alembic', redis_host=redis_host,
                   ttl=schema_maxtime, redis_conn=redis_conn):
            schema_update(db_engine, models, migrate=True,
                          schema_maxtime=schema_maxtime)
    else:
        # just wait for schema update
        schema_update(db_engine, models)


def _init_db(db_url=None, engine=None, connection_timeout=0,
             geo_support=True):
    global db_engine
    if not db_engine:
        try:
            db_engine = engine or create_engine(
                db_url, pool_pre_ping=True,
                pool_recycle=connection_timeout)
        except Exception as ex:
            logging.error('action=init_db status=error message=%s' % str(ex))
            return None
    if db_engine.url.drivername == 'sqlite':
        db_engine.execute('pragma foreign_keys=on')
        if geo_support:
            listen(db_engine, 'connect', _load_spatialite)
    db = get_session(scoped=True)
    if hasattr(db, 'query_property'):
        Base.query = db.query_property()
        db.remove()
    Base.metadata.create_all(bind=db_engine)
    return db_engine


def _load_spatialite(dbapi_conn, connection_record):
    global spatialite_loaded
    for lib in constants.LIB_MOD_SPATIALITE:
        if os.path.exists(lib):
            dbapi_conn.enable_load_extension(True)
            dbapi_conn.load_extension(lib)
            logging.info('action=load_spatialite message=geo_support_added')
            spatialite_loaded = True
            return



# Utilities

import connexion
from datetime import datetime
from flask import g, jsonify
from flask_cors import CORS
from html.parser import HTMLParser
import logging
import random
import string

from . import grants, utils
from .access import AccessControl


def gen_id(length=8, prefix='x-', chars=(
        '-' + string.digits + string.ascii_uppercase + '_' +
        string.ascii_lowercase)):
    """Record IDs are random 48-bit RFC-4648 radix-64 with a fixed prefix
    to make them (somewhat) more human-recognizable

    First 15 bits are generated from Unix epoch to make them sortable
    by date (granularity 24 hours); rolls over after year 2107
    """
    def _int2base(x, chars, base=64):
        return _int2base(x // base, chars, base).lstrip(chars[0]) + chars[
            x % base] if x else chars[0]
    return (prefix +
            _int2base((utcnow() - datetime(2018, 1, 1)).days * 8 +
                      random.randint(0, 8), chars) +
            ''.join(random.choice(chars) for i in range(length - 3)))


def initialize_app(application, config, models):
    """ Initialize the Flask app defined by openapi.yaml

    params:
      application - a connexion object
      config - a flask config object
      models - the SQLalchemy models
      init_func - any other function to call
    """

    logging.basicConfig(level=config.LOG_LEVEL,
                        format='%(asctime)s %(levelname)s %(message)s',
                        datefmt='%m-%d %H:%M:%S')
    logging.getLogger('werkzeug').setLevel(logging.ERROR)
    application.app.config.from_object(config)
    application.add_api(config.OPENAPI_FILE)
    application.add_error_handler(400, utils.render_status_400)
    application.add_error_handler(connexion.ProblemException,
                                  utils.render_problem)
    CORS(application.app,
         resources={r"/api/*": {'origins': config.CORS_ORIGINS}},
         supports_credentials=True)
    AccessControl().load_rbac(config.RBAC_FILE)
    grants.Grants(models).load_defaults(config.DEFAULT_GRANTS)
    logging.info(dict(action='initialize_app', port=config.PORT))
    return application.app


def render_status_400(error):
    return jsonify(dict(
        message=error.description,
        error=dict(status=error.name, code=error.code))), error.code


def render_problem(error):
    return jsonify(dict(
        message=error.detail,
        error=dict(status=error.title, code=error.status))), error.status


def req_duration():
    return '%.3f' % (utcnow().timestamp() -
                     g.request_start_time.timestamp())


def utcnow():
    """ for mocking: unittest.mock can't patch out datetime.utcnow directly """
    return datetime.utcnow()


def strip_tags(html):
    """Convert html to plain-text by stripping tags"""
    s = HtmlStripper()
    s.feed(html)
    return s.get_data()


class HtmlStripper(HTMLParser):
    def __init__(self):
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return ' '.join(self.fed)



"""session_auth

Functions for login, password and role authorization

created 26-mar-2020 by richb@instantlinux.net
"""

from datetime import datetime, timedelta
from flask import g, request
import jwt
import logging
from passlib.hash import sha256_crypt
from sqlalchemy import or_
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.exc import NoResultFound
import string
import time

from .access import AccessControl
from .messaging.confirmation import Confirmation
from . import constants
from .constants import i18n
from .service_registry import ServiceRegistry


class SessionAuth(object):
    def __init__(self, config=None, models=None, func_send=None):
        self.config = config
        self.models = models
        self.jwt_secret = config.JWT_SECRET
        self.login_attempts_max = config.LOGIN_ATTEMPTS_MAX
        self.login_lockout_interval = config.LOGIN_LOCKOUT_INTERVAL
        self.login_session_limit = config.LOGIN_SESSION_LIMIT
        self.login_admin_limit = config.LOGIN_ADMIN_LIMIT
        self.func_send = func_send

    def account_login(self, username, password, roles_from=None):
        """ Log in with username or email

        parameters:
          username - account name or email
          password - credential
          identity_from - model from which to find contact info
          roles_from - model for which to look up authorizations
        """

        try:
            if '@' in username:
                account = g.db.query(self.models.Account).join(
                    self.models.Person).filter(
                        self.models.Person.identity == username,
                        self.models.Account.status == 'active').one()
                username = account.name
            else:
                account = g.db.query(self.models.Account).filter_by(
                    name=username, status='active').one()
        except NoResultFound:
            return dict(username=username, message='not valid'), 403
        except OperationalError as ex:
            logging.error('action=login message=%s' % str(ex))
            return dict(message='DB operational error, try again'), 403
        if (account.invalid_attempts >= self.login_attempts_max and
            account.last_invalid_attempt + timedelta(
                seconds=self.login_lockout_interval) > datetime.utcnow()):
            time.sleep(5)
            return dict(username=username, message='locked out'), 403
        if account.password == '':
            logging.error("username=%s, message='no password'" % username)
            return dict(username=username, message='no password'), 403
        elif not sha256_crypt.verify(password, account.password):
            account.invalid_attempts += 1
            account.last_invalid_attempt = datetime.utcnow()
            logging.info(
                'action=login username=%s credential=invalid attempt=%d' %
                (username, account.invalid_attempts))
            g.db.commit()
            return dict(username=username, message='not valid'), 403
        logging.info('action=login username=%s account_id=%s' %
                     (username, account.id))
        account.last_login = datetime.utcnow()
        account.invalid_attempts = 0
        g.db.commit()
        # connexion doesn't support cookie auth. probably just as well,
        #  this forced me to implement a better JWT design with nonce token
        # https://github.com/zalando/connexion/issues/791
        #  TODO come up with a standard approach, this is still janky
        #  and does not do signature validation (let alone PKI)

        # TODO research concerns raised in:
        # https://paragonie.com/blog/2017/03/jwt-json-web-tokens-is-bad-
        #  standard-that-everyone-should-avoid
        if (account.password_must_change):
            roles = ['person', 'pwchange']
        # TODO parameterize model name contact
        elif (g.db.query(self.models.Contact).filter(
                self.models.Contact.info == account.owner.identity,
                self.models.Contact.type == 'email'
                ).one().status == 'active'):
            roles = ['admin', 'user'] if account.is_admin else ['user']
            if roles_from:
                roles += self.get_roles(account.uid, roles_from)
        else:
            roles = []
        duration = (self.login_admin_limit if account.is_admin
                    else self.login_session_limit)
        ses = g.session.create(account.uid, roles, acc=account.id,
                               identity=account.owner.identity,  ttl=duration)
        if ses:
            retval = dict(
                jwt_token=jwt.encode(
                    ses, self.jwt_secret, algorithm='HS256').decode('utf-8'),
                resources=ServiceRegistry(self.config).find()['url_map'],
                settings_id=account.settings_id)
            if hasattr(account.settings, 'default_storage_id'):
                retval['storage_id'] = account.settings.default_storage_id
            return retval, 201
        else:
            return dict(message='rejected'), 403

    def change_password(self, uid, new_password, reset_token,
                        old_password=None):
        try:
            account = g.db.query(self.models.Account).filter_by(
                uid=uid, status='active').one()
        except NoResultFound:
            return dict(uid=uid, message='account not found'), 404
        logmsg = dict(action='change_password', resource='account',
                      username=account.name, uid=uid)

        if self._password_weak(new_password):
            return dict(message='rejected weak password'), 405
        if old_password:
            if not sha256_crypt.verify(old_password, account.password):
                msg = 'invalid credential'
                logging.warning(dict(message=msg, **logmsg))
                return dict(username=account.name, message=msg), 403
        else:
            retval = Confirmation(self.config, self.models).confirm(
                reset_token)
            if retval[1] != 200:
                msg = 'invalid token'
                logging.warning(dict(message=msg, **logmsg))
                return dict(username=account.name, message=msg), 405
        account.password = sha256_crypt.hash(new_password)
        account.password_must_change = False
        g.db.commit()
        logging.info(dict(message='changed', **logmsg))
        return dict(id=account.id, uid=uid, username=account.name), 200

    def forgot_password(self, identity, username):
        logmsg = dict(action='forgot_password', identity=identity,
                      username=username)
        try:
            account = g.db.query(self.models.Account).join(
                self.models.Person).filter(
                    self.models.Account.is_admin == 0,
                    self.models.Account.status == 'active',
                or_(
                    self.models.Account.name == username,
                    self.models.Person.identity == identity)).one()
            id = g.db.query(self.models.Contact).filter_by(
                type='email', info=account.owner.identity).one().id
        except NoResultFound:
            logging.warning(dict(message='not found', **logmsg))
            return dict(message='username or email not found'), 404
        logging.info(logmsg)
        return Confirmation(self.config, self.models).request(
            id, message=i18n.PASSWORD_RESET, func_send=self.func_send)

    def get_roles(self, uid, member_model, resource=None, id=None):
        """Get roles that match uid / id for a resource
        Each is in the form <resource>-<id>-<privacy level>

        params:
          uid - user ID
          member_model - the DB model that defines membership in resource
          resource - the resource that defines privacy (e.g. list)
          id - ID of the resource (omit if all are desired)
        """
        acc = AccessControl()
        if not resource:
            resource = acc.primary_res
        column = '%s_id' % resource
        try:
            query = g.db.query(member_model).filter_by(uid=uid,
                                                       status='active')
            if id:
                query = query.filter_by(**{column: id})
        except Exception as ex:
            logging.warning('action=get_roles message="%s"' % str(ex))
        roles = []
        for record in query.all():
            if not hasattr(record, resource):
                continue
            # Skip first defined level (always 'public')
            for level in acc.privacy_levels[1:]:
                roles.append('%s-%s-%s' % (resource, getattr(record, column),
                                           level))
                if level == record.authorization:
                    break
        return roles

    def update_auth(self, member_model, id, resource=None, force=False):
        """Check current access, update if recently changed

        params:
          member_model - model (e.g. Guest) which defines membership in
            resource
          id - resource id of parent resource
          resource - parent resource for which membership should be checked
        """
        acc = AccessControl()
        if not resource:
            resource = acc.primary_res
        current = set(acc.auth)
        if force or ('%s-%s-%s' % (resource, id, constants.AUTH_INVITEE)
                     not in current):
            # TODO handle privilege downgrade member/host->invitee
            current |= set(self.get_roles(acc.uid, member_model,
                                          resource=resource, id=id))
            creds = request.authorization
            g.session.update(creds.username, creds.password, 'auth',
                             ':'.join(current))

    @staticmethod
    def _password_weak(password):
        return (not set(password).intersection(string.ascii_lowercase),
                not set(password).intersection(string.ascii_uppercase),
                not set(password).intersection(string.digits),
                not set(password).intersection(string.punctuation)
                ).count(True) > 1


# Support openapi securitySchemes: basic_auth and api_key endpoints

def basic(username, password, required_scopes=None):
    """This is a modified basic-auth validation function. The auth login
    controller method generates a random 8-byte token, stores it in
    the session_manager object as token_auth, and sends it to javascript
    authProvider. The dataProvider must send it back to us as
    basic-auth (base64-encoded).

    Vulnerable to session-hijacking if auth packets aren't encrypted
    end to end, but "good enough" until OAuth2 effort is completed.

    Implemented because of https://github.com/zalando/connexion/issues/791
    """

    session = g.session.get(username, password)
    token_auth = 'ok'
    if session is None or 'jti' not in session:
        token_auth = 'missing'
    elif session['jti'] != password or session['sub'] != username:
        token_auth = 'invalid'
    elif 'exp' in session and datetime.utcnow().strftime(
            '%s') > session['exp']:
        token_auth = 'expired'

    if token_auth != 'ok':
        logging.info('action=logout username=%s token_auth=%s' % (
            username, token_auth))
        if session:
            g.session.delete(username, password)
        return None
    return {'uid': username}


def api_key(token, required_scopes=None):
    logging.info('action=api_key token=%s' % token)
    return {'sub': 'user1', 'scope': ''}
